/*
 * DCAP - Direct CAPture
 * Copyright (C) 2020 Oliver Kirsebom
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <TMath.h>
#include <TF1.h>
#include <DCAP/DirectCapture.h>

using namespace std;
using namespace AUSA::Constants;
using namespace ORM;
using namespace DCAP;


double DCAP::sigmaTheo(double Eg, double E, double Jf, double li, double lf, double R, 
    double Z1, double M1, double jp, double Z2, double M2, double jt) {
    double mu = M1 * M2 / (M1 + M2);
    double w = (2.*Jf+1.) * (2.*li+1.) / ((2.*jp+1.) * (2.*jt+1.) * (2.*lf+1.));
    auto cg = ClebschGordan(li, 1); //j1,j2
    double cgCoeff = cg.Coefficient(0, 0, lf); //m1,m2,J
    double sig = 0.0716 * pow(mu, 1.5) * pow(Z1/M1 - Z2/M2, 2.) 
                * pow(Eg, 3.) / pow(E, 1.5) * w 
                * pow(cgCoeff, 2.) * pow(R, 2.);
    return sig;
}

double DCAP::sigmaExp(double Eg, double E, double Jf, double Tf, vector<double>& liVec, 
    vector<double>& lfVec, vector<vector<double>>& RMat, vector<double>& SVec, 
    double Z1, double M1, double jp, double tp, double mp, double Z2, double M2, 
    double jt, double tt, double mt) {
    double sig = 0;
    for (size_t i=0; i < lfVec.size(); i++) { // loop over lf
        auto& lf = lfVec[i];
        double sig_lf = 0;
        for (size_t j=0; j < liVec.size(); j++) { // loop over li
            auto& li = liVec[j];
            auto& R = RMat[i][j];
            sig_lf += sigmaTheo(Eg, E, Jf, li, lf, R, Z1, M1, jp, Z2, M2, jt);
        }
        auto cg = ClebschGordan(tp, tt); //t1,t2
        double cgCoeffIso = cg.Coefficient(mp, mt, Tf); //m1,m2,T
        auto& S = SVec[i];
        sig += pow(cgCoeffIso, 2.) * S * sig_lf;
    }
    return sig;
}

double DCAP::sigmaExp(double Eg, double E, double Jf, double Tf, double li, double lf, 
    double R, double S, double Z1, double M1, double jp, double tp, double mp, 
    double Z2, double M2, double jt, double tt, double mt) {
        vector<double> liVec{li};
        vector<double> lfVec{lf};
        vector<double> SVec{S};
        vector<vector<double>> RMat{vector<double>{R}};
        return sigmaExp(Eg, E, Jf, Tf, liVec, lfVec, RMat, SVec, Z1, M1, jp, tp, mp, Z2, M2, jt, tt, mt);
}

double DCAP::CMEnergyProjectile(double Ep, double M1, double M2) {
    double EpCM = Ep * TMath::Power(M2/(M1+M2), 2.);
    return EpCM;
}

double DCAP::CMEnergy(double Ep, double M1, double M2) {
    double ECM = Ep * M2/(M1+M2);
    return ECM;
}

double DCAP::GammaEnergy(double Ep, double Exf, double M1, double M2, double Q) {
    double Exi = CMEnergy(Ep,M1,M2) + Q;
    double Eg = Exi - Exf;
    return Eg;
}

double DCAP::radialIntegral(initialWaveFunc& initialWf, finalWaveFunc& finalWf, E1Operator& OE1, double rMax) {
    TF1 f{"", [&](double* x, double* p) { 
        double r = x[0];
        return finalWf(r) * OE1(r) * initialWf(r) * pow(r, 2.);
    },0,0,0};    
    double eps = 0.001;
    return f.Integral(eps, rMax);
}

double DCAP::directCaptureCrossSec(shared_ptr<ParticleChannel>& initial, shared_ptr<ParticleChannel>& final,
    double beamEnergy, double excEnergy, double specFactor, double rMax) {

    auto& pair = initial->getPair();
    auto& p1 = pair.getP0();
    auto& p2 = pair.getP1();

    double initialEnergy = CMEnergy(beamEnergy, p1.getMass(), p2.getMass()); //lab->cm
    double finalEnergy = final->QValue() + excEnergy;
    double gammaEnergy = initialEnergy - finalEnergy;
    
    // keV -> MeV
    double Ei = initialEnergy / 1e3; 
    double Ef = finalEnergy / 1e3; 
    double Eg = gammaEnergy / 1e3; 

    double Jf = final->getSpin(); 
    double Tf = final->getIsospin();

    double Z1 = p1.getZ();
    double M1 = p1.getA();
    double jp = p1.getSpin().getValue();
    double tp = p1.getIsospin().getValue();
    double mp = p1.getIsospinProj();

    double Z2 = p2.getZ();
    double M2 = p2.getA();
    double jt = p2.getSpin().getValue();
    double tt = p2.getIsospin().getValue();
    double mt = p2.getIsospinProj();

    double li = initial->getL().getValue();
    double lf = final->getL().getValue();

    double S = specFactor;

    double R = 0;
    auto cg = ClebschGordan(li, 1);
    double cgCoeff = cg.Coefficient(0, 0, lf);
    if (TMath::Abs(cgCoeff) > 0.) {
        auto finalWf = finalWaveFunc(final, finalEnergy, rMax);
        double V0 = finalWf.getSquareWellDepth();
        auto initialWf = initialWaveFunc(initial, initialEnergy, V0);
        auto OE1 = E1Operator(gammaEnergy);
        R = abs(DCAP::radialIntegral(initialWf, finalWf, OE1, rMax)); //only compute radial integral if CG is non-zero
    }

    double sig = sigmaExp(Eg, Ei, Jf, Tf, li, lf, R, S, Z1, M1, jp, tp, mp, Z2, M2, jt, tt, mt);
    return sig;
}